functor DatasetFun(structure Acl' : ACL
		   structure Throttle : THROTTLE
		   structure AcapTime' : ACAPTIME)
  :> DATASET 
  where type 'a T.throttle = 'a Throttle.throttle 
    and type AcapTime.acaptime = AcapTime'.acaptime 
    and type Acl.acl = Acl'.acl 
    and type Acl.Rights.rights = Acl'.Rights.rights
    and type Acl.Auth.ident = Acl'.Auth.ident =
struct
  structure T = Throttle
  structure V = Value
  structure AcapTime = AcapTime'
  structure Acl = Acl'

  structure SV = SyncVar
  structure MC = Multicast

  type metadata = {acl : Acl.acl option, attribute : string,
		   size : int list, value : V.value}

  type storedata = {acl : Acl.acl option, attribute : string, 
		    value : V.value option}

  type attribute = metadata

  datatype updatemsg =
    NEW of export
  | CHANGE of (export * export)
  | DELETE of export

  and dataset = 
    DSET of 
    {data : entry list, multichan : updatemsg MC.mchan,
     inherits : (string option * dataset option), 
     dielistener : unit SV.ivar option,
     null : entry} SV.mvar
  | EMPTY

  and subdataattr = 
    NoSub
  | Sub of dataset
  | Elsewhere of string list

  and state =
    Normal
  | Default
  | Nil 
	
  withtype entry = {name : attribute, subdataset : subdataattr,
		    modtime : AcapTime.acaptime * attribute, 
		    attrs : attribute list, 
		    iattrs : attribute list,
		    isubdataset : subdataattr,
		    state : state}

  and export = entry 
    * (attribute * Acl.Auth.ident * Acl.Rights.rights -> bool)

  exception EntryNameNotSingleValue
  exception EmptyDataset
  exception InconsistentInheritance

  (* ENTRY FUNCTIONS *)
  exception EnforcedValidation of (string * string)
  exception Modified of AcapTime.acaptime
  exception SyntaxError of string
  exception DoubleAttribute
  exception DatasetRefer of string list
  exception RenameEntry of string
  exception NotPermitted
  exception NoAcl

  (* syntax checks! *)
  fun legalEntryChar (c, acc) = acc andalso c <> #"/" 
    
  fun legalEntryName s = 
    case Substring.getc (Substring.all s) of
      NONE => true
    | (SOME (c, substr)) => c <> #"/" andalso 
	Substring.foldl legalEntryChar true substr

  fun legalDatasetName s =
    case String.fields (fn c => c = #"/") s of
      nil => false
    | (r::path) => if r <> "" then false else List.all legalEntryName path

  (* acl checks ! *)
  fun canDo (null as {name, modtime, state,
		      attrs, subdataset, 
		      iattrs, isubdataset} : entry)
            (attr : attribute, ident, right) =
    let
      val attracl = ("dataset.acl." ^ (#attribute attr))

      fun findacl (nil, NONE) = raise NoAcl
	| findacl (nil, (SOME acl)) = acl
	| findacl (({attribute,value,...} : attribute)::l, NONE) =
	if attribute = "dataset.acl" then
	  findacl (l, (SOME (Acl.fromValue value)))
	else if attribute = attracl then Acl.fromValue value
	     else findacl (l, NONE)
	| findacl ({attribute,value,...}::l, d as (SOME _)) =
	if attribute = attracl then Acl.fromValue value
	else findacl (l, d)

      val acl = case #acl attr of
	NONE => findacl (attrs, NONE)
     | (SOME acl) => acl
    in
      Acl.Rights.has (Acl.myrights ident acl, right)
    end

  fun canRead f ident attr = f (attr, ident, Acl.Rights.ACLr)
  fun canWrite f ident attr = f (attr, ident, Acl.Rights.ACLw)
  fun canSearch f ident attr = f (attr, ident, Acl.Rights.ACLx)
  fun canInsert f ident attr = f (attr, ident, Acl.Rights.ACLi)
  fun canAdmin f ident attr = f (attr, ident, Acl.Rights.ACLa)

  local
    (* create null attributes *)
    val nullMeta = {acl=NONE, attribute="", size=nil, value=V.Nil}
    fun makeNull attr = {acl=NONE, attribute=attr, size=nil, value=V.Nil}

    fun computeSize (V.Single s) = [String.size s]
      | computeSize (V.List l) = map String.size l
      | computeSize V.Nil = nil
      | computeSize V.Default = nil

    exception ChangeName of string
    exception DefaultEntry
    exception DeleteEntry
    exception CreateSubdataset of (string list * Acl.acl option)
    exception DeleteSubdataset
    exception MyEnforcedVal of string
    exception NoExn

    fun storeToMeta (ident, aclf,
		     ({acl, attribute, value} : storedata),
		     (old as {acl=oldAcl, attribute=oldAttr, 
			      size=oldSize, value=oldValue} : metadata)) =
      let 
	val canWrite = canWrite aclf ident
	val canAdmin = canAdmin aclf ident

	val _ = if attribute = "modtime" then 
	  raise (MyEnforcedVal "modtime")
		else ()

	val _ = if isSome value andalso not (canWrite old)
		  then (print ("notpermitted " ^ attribute ^ "\n");
			raise NotPermitted) else ()

	val _ = if isSome acl andalso not (canAdmin old)
		  then raise NotPermitted else ()
		  
	(* are we attempting to delete or rename this entry? *)
	val _ = if attribute = "entry" then
	  case value of
	    NONE => ()
	  | (SOME (V.List _)) => raise 
	      (MyEnforcedVal "entry")
	  | (SOME (V.Single n)) => raise (ChangeName n)
	  | (SOME (V.Default)) => raise (DefaultEntry)
	  | (SOME (V.Nil)) => raise (DeleteEntry)
		else ()
		  
	val _ = if attribute = "subdataset" then
	  case (value, acl) of
	    (NONE, _) => ()
	  | (SOME (V.Single _), _) => raise (MyEnforcedVal "subdataset")
	  | (SOME (V.List l), acl) => raise (CreateSubdataset (l, acl))
	  | (SOME V.Nil, _) => raise DeleteSubdataset
	  | (SOME V.Default, _) => raise DeleteSubdataset (* XXXXXX *)
		else ()
		  
	val acl = if isSome acl then acl else oldAcl
	val (value, size) = if isSome value then (valOf value, 
						  computeSize (valOf value))
			    else (oldValue, oldSize)
      in
	{acl=acl, attribute=attribute, size=size, value=value}
      end
    
      fun validateSubdata nil = ()
	| validateSubdata (a::l) = 
	   if a = "." then raise MyEnforcedVal "subdata"
	   else validateSubdata l

      (* fetch and search have to return "entry" and "modtime" attributes! *)
      fun subdataset' (NoSub) = makeNull "subdataset"
	| subdataset' (Sub _) =
	{acl=NONE, attribute="subdataset", size=[1], value = V.List ["."]}
	| subdataset' (Elsewhere l) =
	{acl=NONE, attribute="subdataset", size=map String.size l,
	 value = V.List l}

      val getName : attribute -> string = #attribute
	
  in
    (* store' : ident -> entry -> entry -> storedata list -> entry option *)
    fun store' (ident, nullE : entry) = 
      let
	val aclf = canDo nullE

	fun st entry nil = (SOME entry)
	  | st {state,attrs,name,modtime,subdataset,iattrs,isubdataset} 
	  ((sd as {attribute=tochange,...})::l) = 
      let
	val (a, others) = (List.partition 
			   (fn ({attribute=a,...}) => (a = tochange)) attrs)
	  
	val (attr, exn) = (((case a of
			       nil => [storeToMeta(ident, aclf, sd, nullMeta)]
			     | [a] => [storeToMeta(ident, aclf, sd, a)]
			     | _ => raise DoubleAttribute), NoExn)
			   handle exn => (a, exn))
      in
	case exn of
	  NoExn => st {state=Normal, attrs=attr @ others, iattrs=nil,
		       name=name, modtime=modtime,
		       subdataset=subdataset, isubdataset=NoSub} l
	| ChangeName n => st {state=state, attrs=attrs,
			      name={acl=NONE, attribute="entry",
				    size=[String.size n], 
				    value=V.Single n},
			      modtime=modtime,
			      subdataset=subdataset,
			      iattrs=nil, isubdataset=NoSub} l
	| DefaultEntry => 
	  if List.null l then 
	    ((case subdataset of
		(NoSub | Elsewhere _) => ()
	      | (Sub a) => dsubdset (ident, nullE) a);
		NONE)
	  else raise (SyntaxError "trying to delete and store")
	| DeleteEntry =>
	    if List.null l then
	      SOME {state=Nil, 
		    attrs=nil, iattrs=nil,
		    subdataset=(case subdataset of
				  (NoSub | Elsewhere _) => NoSub
				| (Sub a) => (dsubdset (ident, nullE) a; 
					      NoSub)),
		    isubdataset=NoSub,
		    modtime=modtime, name=name}
	    else raise (SyntaxError "trying to delete and store")
        | MyEnforcedVal a => raise EnforcedValidation (getName name, a)
        | CreateSubdataset (["."], acl) => 
	    (case subdataset of
	       (Sub subd) => st {state=state, 
				 attrs=attrs,
				 name=name,
				 modtime=modtime,
				 subdataset=subdataset, 
				 isubdataset=subdataset,
				 iattrs=nil} l
	     | _ => let val sub = Sub (csubdset (ident, nullE))
			val _ = print "got CreateSubdataset (1)\n" in
		 st {state=state, attrs=attrs, 
		     name=name, modtime=modtime,
		     subdataset=sub, isubdataset=sub,
		     iattrs=nil} l end)
	| CreateSubdataset (subl, acl) => 
	       (validateSubdata subl;
		(case subdataset of
		   (Sub subd) => (dsubdset (ident, nullE) subd)
		 | _ => ());
		   st {state=state, attrs=attrs, 
		       name=name, modtime=modtime,
		       subdataset=Elsewhere subl,
		       isubdataset=Elsewhere subl,
		       iattrs=nil} l)
	| DeleteSubdataset => 
	       (case subdataset of
		  (NoSub | Elsewhere _) => 
		    st {state=state, attrs=attrs, 
			name=name, modtime=modtime,
			subdataset=subdataset, iattrs=nil,
			isubdataset=subdataset} l
		| (Sub a) => 	(* delete subdataset *)
		    (dsubdset (ident, nullE) a;
		     st {state=state, attrs=attrs, 
			 name=name, modtime=modtime,
			 subdataset=NoSub, 
			 iattrs=nil, isubdataset=NoSub} l))
        | exn => raise exn
      end in st end
	
    (* given an inherited entry and an entry, create entry's iattrs *)
    and syncEntry (e as {state=Nil,...}, _) = e
      | syncEntry (e as {state=Default, name, modtime,...}, NONE) = 
      {state=Default, name=name, modtime=modtime, 
       attrs=nil, subdataset=NoSub,
       iattrs=nil, isubdataset=NoSub}
      | syncEntry ({state, name, modtime, attrs, subdataset,
		    iattrs, isubdataset},
		   SOME({iattrs=inh, isubdataset=isub,...} : entry)) =
      let
	(* adds the inherited entries from l to al *)
	fun addem (al, nil) = al
	  | addem (al, (a as {attribute=an,...} : metadata)::l) =
	  if List.exists (fn {attribute,...} => attribute=an) al then
	    addem (al, l)
	  else
	    addem (a::al, l)

	val isub = case isubdataset of
	  NoSub => isub
	| _ => isubdataset
      in
	{state=state,
	 name=name, modtime=modtime, 
	 attrs=attrs, subdataset=subdataset,
	 iattrs=(addem (attrs, inh)), isubdataset=isub} : entry
      end
      | syncEntry ({state, name, modtime,
		    attrs, subdataset,
		    iattrs, isubdataset}, NONE) =
      {state=state, name=name, modtime=modtime,
       attrs=attrs, subdataset=subdataset,
       iattrs=attrs, isubdataset=subdataset}

    (* creates an entry, given initial data and inherited entry *)
    (* val create : (ident * entry) -> string -> storedata list -> 
                           entry option -> time -> entry*)
    and createEntry (ident, nullE : entry) myname sd 
                    (ientry : entry option) time = 
      let
	val _ = if legalEntryName myname then () 
		else raise (SyntaxError "bad entry name")
	
	val aclf = canDo nullE

	fun create' e nil = e
	  | create' (e as {state, name, modtime,
			   attrs,subdataset,
			   iattrs, isubdataset}) (sd::l) =
	  ((create' {state=state, name = name, modtime = modtime, 
		     attrs=(storeToMeta(ident, aclf, sd, nullMeta))::attrs,
		     subdataset=subdataset, 
		     iattrs=iattrs, isubdataset=isubdataset} l)
	   handle CreateSubdataset(["."], acl) => 
	          let val _ = print "got CreateSubdataset (2)\n"
		      val sub = Sub (csubdset (ident, nullE)) in
		    create' {state=state, name=name, modtime=modtime,
			     attrs=attrs, subdataset=sub,
			     isubdataset=sub, iattrs=iattrs} l end
		| CreateSubdataset (subl, acl) => (validateSubdata subl;
               	    create' {state=state, name=name, modtime=modtime,
			     attrs=attrs, subdataset=Elsewhere subl,
			     iattrs=iattrs, isubdataset=isubdataset} l)
		| MyEnforcedVal a => raise EnforcedValidation (myname, a)
		| SyntaxError msg => raise (SyntaxError msg)
		| Acl.SyntaxError msg => raise (SyntaxError
						("ACL: " ^ msg))
		| exn => raise (SyntaxError ("this create made no sense: " ^
					     (exnMessage exn))))

	val newentry = create' {state=Normal,
				name={acl=NONE,
				      attribute="entry",
				      size=[String.size myname],
				      value=V.Single myname}, 
				modtime=modtimeNow (),
				attrs=nil, subdataset=NoSub,
				iattrs=nil, isubdataset=NoSub} sd
      in
	syncEntry (newentry, ientry)
      end
    
    (* val storeEntry : (ident * null : entry) -> entry -> entry option -> 
           storedata list -> time option -> entry option *)
    and storeEntry (ident, nullE : entry) (entry as 
					   {state, name, modtime,
					    attrs, subdataset,
					    iattrs, isubdataset} : entry) 
                                          ientry sd time =
      let
	val (actime, _) = modtime

	val _ = if isSome time andalso
	  AcapTime.compare (actime, valOf time) = GREATER
	  then raise (Modified actime)
	else ()

	val newentry = store' (ident, nullE)
	                      {state=state, name=name, modtime=modtimeNow (),
			       attrs=attrs, subdataset=subdataset,
			       iattrs=nil, isubdataset=NoSub} sd
      in
	case (newentry, ientry) of
	  (NONE, NONE) => NONE
	| (NONE, SOME _) => SOME(syncEntry ({state=Default,
					     name=name, 
					     modtime=modtimeNow (),
					     attrs=nil, iattrs=nil,
					     subdataset=NoSub,
					     isubdataset=NoSub}, ientry))
	| (SOME newentry, _) => SOME(syncEntry (newentry, ientry))
      end

    and csubdset (ident, null) =
      let
	val d = SV.mVarInit {data=nil, multichan=MC.mChannel (),
			     inherits=(NONE, NONE), dielistener=NONE,
			     null=null}
      in
	DSET(d)
      end
    
    and dsubdset (ident, null) (DSET dset) =
      let 
	val ({data,multichan,...}) = SyncVar.mTake dset

	val aclf = canDo null

	fun dentry (e : entry) =
	  (ignore (storeEntry (ident, null) e NONE [{attribute="entry",
						     value=SOME (V.Nil),
						     acl=NONE}]);
	   MC.multicast (multichan, DELETE(e, aclf)))
      in
	app dentry data
      end
      | dsubdset _ EMPTY = raise EmptyDataset

    and modtimeNow () = 
      let 
	val t = AcapTime.now ()
	val s = AcapTime.toString t
      in (t, {acl=NONE, attribute="modtime", size=[String.size s],
	      value=V.Single s} : attribute)
      end

    
    (* gets the attributes named *)
    (* val fetch : entry -> string -> attribute *)
    fun fetchEntry (ident, doinh, (e as {state, name, modtime=(_,modtime),
					 attrs, subdataset, 
					 iattrs, isubdataset}, aclf)) an =
      let
      val attrs = if doinh then iattrs else attrs
      val subdataset = if doinh then isubdataset else subdataset
      val canRead = canRead aclf ident

      val attr = if an = "entry" then SOME name
		 else if an = "modtime" then SOME (modtime)
		      else if an = "subdataset" then 
			     SOME (subdataset' subdataset)
			   else List.find (fn ({attribute,...}:attribute) =>
					   an = attribute) attrs
			     
      val attr = case attr of
	         (SOME x) => if canRead x then x
			     else makeNull an
	       | NONE => makeNull an
    in
      attr
    end

    (* gets the matching attributes *)
    (* val search : bool -> entry -> (string -> bool) -> attribute list *)
    fun searchEntry (ident, doinh, (e as {state, name,modtime=(_,modtime),
					  attrs, subdataset,
					  iattrs, isubdataset}, aclf)) f =
      let
	val attrs = if doinh then iattrs else attrs
	val subdataset = if doinh then isubdataset else subdataset
	val canRead = canRead aclf ident

	val attrs = List.filter (fn (a as {attribute,...}) => 
			f attribute andalso canRead a) attrs

	(* now we figure out if "entry" or "modtime" match *)
	val attrs = if f "entry" andalso canRead name
		      then name::attrs else attrs

	val attrs = if f "modtime" andalso canRead modtime
		      then modtime::attrs else attrs
		      
	fun g NoSub = false
	  | g _ = true

	val attrs = if (g subdataset) andalso (f "subdataset")
		    then (subdataset' subdataset)::attrs else attrs
      in
	attrs
      end

    fun expname ({name=(name as {value=(V.Single n),...}),...} : entry, 
		 aclf) = n
      | expname _ = raise EntryNameNotSingleValue

    fun name ({name=(name as {value=(V.Single n),...}),...} : entry) = n
      | name _ = raise EntryNameNotSingleValue

    (* get the entry's name *)
    (* val name : export -> name *)
    fun nameEntry (ident, ({name=(name as {value=(V.Single n),...}),...}
                                     : entry, 
			   aclf)) = 
      let
	val canSearch = canSearch aclf ident
      in
	if canSearch name then n else raise NotPermitted
      end
      | nameEntry _ = raise EntryNameNotSingleValue
      
    (* return the time of last modification *)
    (* val gettime : entry -> acaptime *)
    fun timeEntry ({modtime=(modtime,_),...} : entry, aclf) = modtime

      
    fun getsubdset false ({subdataset=Sub subdataset,...} : entry) = 
                   SOME subdataset
      | getsubdset false ({subdataset=Elsewhere l,...}) = raise DatasetRefer l
      | getsubdset true  ({isubdataset=Sub subdataset,...}) = SOME subdataset
      | getsubdset true  ({isubdataset=Elsewhere l,...}) = raise DatasetRefer l
      | getsubdset _ _ = NONE

    fun getsubdset' false ({subdataset=Sub subdataset,...}:entry) = 
                   SOME subdataset
      | getsubdset' true  ({isubdataset=Sub subdataset,...}) = SOME subdataset
      | getsubdset' _ _ = NONE
  end

  fun blankNull () = 
    let
      val a = "$anyone\trwxia";

      val attr = {acl=NONE, 
		  attribute="dataset.acl",
		  value=V.List [a],
		  size=[String.size a]}

      val l = [attr]
    in
      ({name={acl=SOME (Acl.fromString "$anyone\trwxia"), 
	      attribute="name", size=[0],
	      value=V.Single ""}, 
	state=Normal, modtime=modtimeNow (),
	attrs=l, subdataset=NoSub, 
	iattrs=l, isubdataset=NoSub} : entry)
    end
    
  local 
    (* root dataset *)
    fun makeroot () = DSET(SV.mVarInit {data=nil, multichan=MC.mChannel (),
 			inherits=(NONE, NONE), dielistener=NONE, 
			null=blankNull ()})

    val empty = EMPTY
      
    fun isempty EMPTY = true
      | isempty _ = false
  in
    val root = ref empty
    fun init () = root := makeroot ()
  end

  exception NoSuchDataset
  exception AlreadyExists of dataset

  exception DoubleEntry

  fun splitit name =
    case String.fields (fn c => c = #"/") name of
      nil => raise NoSuchDataset
    | (r::dsetpath) => if r <> "" then raise NoSuchDataset else dsetpath
  
  (* produces a blank entry given one it will be inheriting *)
  fun blankEntry ({state, name, modtime, 
		   attrs, subdataset, 
		   iattrs, isubdataset}) =
    {state=Default, name=name, modtime=modtime, 
     attrs=nil, subdataset=NoSub, 
     iattrs=iattrs, isubdataset=isubdataset}

  (* redoInh : (entry list * entry list) -> entry list *)
  fun redoInh (data, idata) =
    let
      fun order (e1, e2) = (name e1) > (name e2)
      val sort = ListMergeSort.sort order
      val data = sort data
      val idata = sort idata

      (* if we have a correspondence, run it through syncEntry.
         otherwise, if it's a purely inherited entry make up a blank
         one.  if it's purely local, syncEntry with no inherited entry *)
      fun redoInh' (nil, nil, acc) = acc
	| redoInh' (e1::data, nil, acc) =
	redoInh' (data, nil, (syncEntry (e1, NONE))::acc)
	| redoInh' (nil, e2::idata, acc) =
	redoInh' (nil, idata, (blankEntry e2)::acc)
	| redoInh' (e1::data, e2::idata, acc) = 
	let
	  val n1 = name e1
	  val n2 = name e2
	in
	  case String.compare (n1, n2) of
	    GREATER => redoInh' (e1::data, idata, (blankEntry e2)::acc)
          | EQUAL => redoInh' (data, idata, (syncEntry (e1, SOME e2))::acc)
	  | LESSER => redoInh' (data, e2::idata, (syncEntry (e1, NONE))::acc)
	end
    in
      redoInh' (data, idata, nil)
    end

  (* listenInh : dataset * updatemsg Multicast.port -> pleaseDie (ivar) *)
  fun listenInh (DSET dset, chan) = 
    let
      val pleaseDie : unit SyncVar.ivar = SyncVar.iVar () 
      val acceptE = MC.recvEvt chan
      val die = SyncVar.iGetEvt pleaseDie
      val takeE = SyncVar.mTakeEvt dset

      fun evt v =
	let
	  val _ = print "Propogating change...\n"

	  val (d as {data,multichan,inherits,dielistener,null}) = 
	                     CML.select[CML.wrap(takeE, fn x => x),
					CML.wrap(die, CML.exit)]

	  val n = case v of (NEW e) => expname e
	                  | (CHANGE (e, ne)) => expname e
			  | (DELETE (e)) => expname e

	  val (oldentry, others) = List.partition (fn e => (name e) = n) data

	  val newentry =
	    case (oldentry, v) of
	      (nil, NEW (e, _)) => blankEntry e
	    | (nil, CHANGE (e, ne)) => raise InconsistentInheritance
	    | (nil, DELETE e) => raise InconsistentInheritance
	    | ([oe], NEW (e, _)) => syncEntry (oe, SOME e)
	    | ([oe], CHANGE (e, ne)) => 
		if expname e = expname ne then syncEntry (oe, SOME (#1 ne))
		else syncEntry(oe, NONE)
	    | ([oe], DELETE e) => syncEntry (oe, NONE)
	    | (_, _) => raise DoubleEntry

	  val aclf = canDo null

	  val others = case v of
	      (NEW _) => (MC.multicast (multichan, (NEW (newentry, aclf)));
			  others)
	    | (CHANGE (e, ne)) => if expname e = expname ne then others else 
	     let 
	       val n = expname ne 
	       val (oldne, others) = 
		     List.partition (fn e => (name e) = n) others

	       (* ne is an export; we just need the entry ! *)
	       val neEntry = #1 ne

	       val ne' = 
		 case oldne of
		   nil => (blankEntry neEntry)
		 | [oldne] => (syncEntry (oldne, SOME (neEntry)))
		 | _ => raise DoubleEntry
	     in
	       MC.multicast (multichan, CHANGE((List.hd oldentry, aclf), 
					       (newentry, aclf)));
	       MC.multicast (multichan, CHANGE((List.hd oldne, aclf), 
					       (ne', aclf)));
	       ne'::others
	     end
	    | (DELETE e) => (MC.multicast (multichan, 
					   CHANGE((List.hd oldentry, aclf), 
						  (newentry, aclf)));
			     others)

	  val data = newentry::others
	in
	  SyncVar.mPut (dset, {data=data, multichan=multichan,
			       inherits=inherits, dielistener=dielistener,
			       null=null})
	end

      fun listen () =
	(CML.select[CML.wrap(acceptE, evt),
		    CML.wrap(die, CML.exit)];
	 listen ())
    in
      CML.spawn listen;
      pleaseDie
    end
    | listenInh (EMPTY, _) = raise EmptyDataset

  (* get the named dataset relative to this one *)
  (* val fetch : dataset -> string list -> dataset *)
  fun fetch' d (nil | (""::nil)) = d
    | fetch' (DSET dset) (first::rest) =
    let
      val ({data,...}) = SyncVar.mGet dset
    in
      (* now, is first a subdataset of ours? *)
      case List.find (fn x => (name x) = first) data of
	NONE => raise NoSuchDataset
      | SOME(x) => (case getsubdset false x of
		      (SOME x) => fetch' x rest
		    | NONE => raise NoSuchDataset)
    end
    | fetch' EMPTY _ = raise NoSuchDataset

  fun fetch (nil | (""::nil)) = !root
    | fetch l = fetch' (!root) l

  (* we're currently traversing this dataset, which is "herePath".
     the path to go is the third param; the depth is the 4th.  "win"
     takes the path & the acl function for that dataset and the entry
     under consideration, and sends out the entry if it wins.  msg is
     the context, if any *)
  fun search' doinh (DSET dset) herePath (nil | (""::nil)) n win msg =
    let (* search this dataset *)
      val (v as {data,multichan,null,...}) = SyncVar.mTake dset

      (* set up the context, if any *)
      val _ = (case msg of NONE => () 
                 | (SOME msg) => msg (herePath, MC.port multichan))
	
      (* put it back *)
      val _ = SyncVar.mPut (dset, v)

      val aclf = canDo null

      (* search for the data and send it out *)
      val _ = List.app (win (herePath, aclf)) data

      (* search any subdatasets *)
      val _ = (case n of 1 => ()
                       | n => 
		 let
		   fun getit e = 
		     (case (getsubdset' doinh e) of
			NONE => NONE
		      | (SOME sd) => (SOME (name e, sd)))

		   val subdsets = List.mapPartial getit data
		 in
		   app (fn (name, d) => search' doinh d 
			                        (herePath ^ "/" ^ name) 
			                        nil (n-1) win msg) 
		       subdsets
		 end)
    in
      ()
    end
  | search' doinh (DSET dset) herePath (first::rest) n win msg =
    let (* go to the right dataset *)
      val (v as {data,...}) = SyncVar.mGet dset
      
      val sub = Option.mapPartial (getsubdset doinh)
	(List.find (fn x => (name x) = first) data)
    in
      case sub of
	NONE => raise NoSuchDataset
      | (SOME d) => search' doinh d (herePath ^ "/" ^ first) rest n win msg
    end
    | search' _ EMPTY _ _ _ _ _ = raise NoSuchDataset

  fun search ident path n sfun msg out doinh =
    let
      val out' = T.call out
      val checkstate = if doinh then fn (e : entry) => #state e <> Nil
		       else fn (e : entry) => #state e = Normal
      fun win (p, aclf) e =
	let val exp = (e, aclf) in	  
	  if checkstate e andalso sfun exp then out' (p, exp) 
	  else ()
	end

      val _ = search' doinh (!root) "" path n win msg
    in
      T.done out
    end
	
  (* store the list of entry/attribute pairs in this dataset *)

  fun store' _ _ dset _ nil _ = dset
    | store' ident ptr {data, multichan,
			inherits=(dsS, dsP), dielistener,
			null} id (("",sdata)::l) time =
    let (* this is the null entry, which contains metadata about the dataset *)
	      (* note to self: dsS = dataset string
	                       dsP = dataset pointer *)
      val newentry = storeEntry (ident, null) null NONE sdata time
	
      (* congrats, we _always_ have a null entry! *)
      val newentry = case newentry of
	            (SOME x) => x
		  | NONE => blankNull ()

      val aclf = canDo newentry

      (* if we changed inheritance, we have to renormalize the entire 
         dataset! *)
      val newin = 
	case #value (fetchEntry (Acl.Auth.master, false, (newentry, aclf))
                                "dataset.inherit") of
	  V.Single d => if legalDatasetName d then SOME d
			else raise EnforcedValidation ("", "dataset.inherit")
	| (V.Nil | V.Default) => NONE
	| _ => raise EnforcedValidation ("", "dataset.inherit")

      (* try to fetch the dataset *)
      (* major bug here; might deadlock!!! *)
      val dsP = 
	case newin of
	  NONE => NONE
	| (newin as (SOME x)) => if newin = dsS then dsP 
				 else SOME (fetch (splitit x))
				   handle NoSuchDataset => NONE

      val _ = print "deciding on new data...\n"

      val (data, dielistener, id) = 
	if newin = dsS then (data, dielistener, id)
	else (* fuck - we have to re-inherit all of our data *)
	  (case dielistener of (* stop listening to whoever we are now *)
	     (SOME dl) => SyncVar.iPut(dl, ())
	   | NONE => ();
	   case dsP of
	     (SOME (DSET dsP)) => (* new inherited dataset *)
	       let
		 val (ds as {data=inheriteddata,
			     multichan=ichan,...}) = SyncVar.mTake dsP

		 val data = redoInh (data, inheriteddata)

		 (* and let's listen to this thing *)
		 val dielistener = SOME (listenInh (ptr, MC.port ichan))
	       in
		 SyncVar.mPut (dsP, ds);
		 (data, dielistener, SOME ds)
	       end
	   | (SOME EMPTY) => raise EmptyDataset
	   | NONE => (* clear out all inheritance *)
	       (print " clearing out inheritance\n";
		(redoInh (data, nil), NONE, NONE)))

      val _ = print "done\n"
    in
      store' ident ptr {data=data,multichan=multichan,
			inherits=(newin, dsP), dielistener=dielistener,
			null=newentry} id l time
    end
    | store' ident ptr {data, multichan,
			inherits, dielistener, null} id 
             ((ename,sdata)::l) time =
    let
      val (e, others) = (List.partition (fn (e) => (name e = ename)) data)

      val idata = case id of NONE => nil | (SOME id) => #data id
      val (ie, _) = (List.partition (fn (e) => (name e = ename)) idata)
	
      val ie =
	case (ie) of
	  nil => NONE
	| [e] => SOME e
	| _ => raise DoubleEntry
	
      val newentry =
 	case (e) of
	  (nil) => SOME(createEntry (ident, null) (ename) 
 			     sdata ie (AcapTime.now ()))
	| [e] => storeEntry (ident, null) e ie sdata time
	| _ => raise DoubleEntry

      val newname = Option.map name newentry

      val _ = case newname of
         	(SOME newname) => 
		  if ename <> newname andalso 
		    isSome (List.find (fn e => name e = newname) data)
		  then (* we're trying to rename it on top of a new name! *)
		    raise (RenameEntry newname)
		  else ()
	      | NONE => ()
		   
      val aclf = canDo null

      val _ =
	case e of (* send new entry notification *) 
	  nil => MC.multicast (multichan, NEW(valOf newentry, aclf))
 	| [e] => (case newentry of (* send change/delete notification *)
		    (SOME newentry) => MC.multicast (multichan, 
				    CHANGE((e, aclf), (newentry, aclf)))
		  | NONE => MC.multicast (multichan, DELETE(e, aclf)))
	| _ => raise DoubleEntry
    in 
      case newentry of
	(SOME newentry) => store' ident ptr {data=newentry::others, 
					     multichan=multichan,
					     inherits=inherits, 
					     dielistener=dielistener,
					     null=null} id l time
      | NONE => store' ident ptr {data=others, multichan=multichan,
				  inherits=inherits, dielistener=dielistener,
				  null=null} id l time
    end

  (* create the named dataset relative to this one; return the new dataset *)
  (* val create : dataset -> string -> time option -> dataset *)
  fun create' ident (DSET dset) (nil | (""::nil)) = (DSET dset)
    | create' ident (ptr as DSET dset) (first::rest) =
    let
      val (v as {multichan, data, null,
		 inherits=(dsS, dsP), dielistener}) = SV.mTake dset

      (* create it here! *)
      val sub = ((Option.mapPartial (getsubdset false)
		  (List.find (fn x => (name x) = first) data))
		 handle exn => (SV.mPut (dset, v);
				raise exn))

      (* if the dataset already exists, sub points to it *)
      val (v as {multichan, data, null,
		 inherits, dielistener}) = 
	case sub of
	  (SOME d) => v
	| NONE => 
	    let
	      val id = case dsP of 
		NONE => NONE
	      | (SOME (DSET id)) => SOME (SyncVar.mGet id)
	      | _ => raise EmptyDataset
	    in
	      store' ident ptr v id
	                   [(first, [{attribute="subdataset",
				      acl=NONE,
				      value=(SOME(V.List ["."]))}])] NONE
	    end

      (* if it didn't already exist, we have to go search for it *)
      val sub = case sub of
	  NONE => (valOf o (getsubdset false) o valOf)
	    (List.find (fn x => (name x) = first) data)
	| (SOME x) => x
    in
      SV.mPut (dset, {multichan=multichan,
		      data=data,
		      null=null,
		      inherits=inherits,
		      dielistener=dielistener});
      if rest = nil then sub else create' ident sub rest
    end
    | create' _ EMPTY _ = raise EmptyDataset

  fun create ident a = create' ident (!root) a

  fun store ident path create nil time = ()
    | store ident path create l time =
    let
      val ptr = if create then create' ident (!root) path
		else fetch path

      val dset = case ptr of (DSET dset) => dset
                           | EMPTY => raise EmptyDataset

      val (v as {inherits=(dsS, dsP),...}) = SyncVar.mTake dset

      val id = 
	case dsP of
	  NONE => NONE
	| (SOME (DSET id)) => SOME (SyncVar.mGet id)
	| _ => raise EmptyDataset

      val v = (store' ident ptr v id l time 
	       handle exn => (SyncVar.mPut (dset, v);
			      raise exn))
      (* handles EnforcedValidation, Modified, SyntaxError, and RenameEntry *)

      val _ = SyncVar.mPut (dset, v)
    in
      ()
    end

  (* acl modification stuff *)
  type aclobj = {dataset : string list,
		 attribute : string option,
		 entry : string option}

  datatype aclmod = 
    SETACL of string * Acl.Rights.rights
  | DELACL of string option
  | GETACL

  exception NotImplemented
    
  (* val modifyacl : Acl.auth.ident -> (aclobj * aclmod) -> Acl.acl *)
  fun modifyacl ident ({dataset=dset, 
			attribute=a, 
			entry=NONE}, operation) =
    let (* fetch the acl *)
      val attracl = 
	case a of NONE => "dataset.acl"
                | (SOME a) => "dataset.acl." ^ a

      val d = case fetch dset of (DSET d) => d
                               | EMPTY => raise EmptyDataset

      val (v as {data,multichan,null,inherits,dielistener}) = SyncVar.mTake d

      val ({name, modtime, state,
	    attrs, subdataset,
	    iattrs, isubdataset}) = null

      fun findacl (nil, NONE) = raise NoAcl
	| findacl (nil, (SOME acl)) = acl
	| findacl (({attribute,value,...} : attribute)::l, NONE) =
	if attribute = "dataset.acl" then
	  findacl (l, (SOME (Acl.fromValue value)))
	else if attribute = attracl then Acl.fromValue value
	     else findacl (l, NONE)
	| findacl ({attribute,value,...}::l, d as (SOME _)) =
	if attribute = attracl then Acl.fromValue value
	else findacl (l, d)

      val _ = if not (Acl.Rights.has 
		      (Acl.myrights ident (findacl (attrs, NONE)),
		       Acl.Rights.ACLa))
		then raise NotPermitted
	      else ()

      val (value, others) =
	case List.partition (fn ({attribute,...}:attribute) => 
			attribute = attracl) attrs of
	  (nil, others) => (V.Nil, others)
	| ([{value,...}], others) => (value, others)
	| _ => raise DoubleAttribute

      val acl = Acl.fromValue value

      val acl = 
	case operation of
	  (SETACL x) => Acl.update acl x
	| (DELACL NONE) => if isSome a then Acl.empty
			   else raise (SyntaxError "can't delete default acl")
	| (DELACL (SOME id)) => Acl.update acl (id, Acl.Rights.empty)
	| GETACL => acl

      val acl' = Acl.toValue acl

      val attrs = ({acl=NONE, attribute=attracl, 
		    size=V.size acl', value=acl'}::others)

      val null = ({name=name, modtime=modtime, state=state,
		   attrs=attrs, subdataset=subdataset,
		   iattrs=attrs, isubdataset=isubdataset})
    in
      SyncVar.mPut (d, {data=data, multichan=multichan, null=null,
			inherits=inherits, dielistener=dielistener});
      acl
    end
    | modifyacl ident ({dataset=dset,
			attribute=a,
			entry=(SOME ename)}, operation) =
    let (* this case modifies the on entry attribute acl *)
      val attr = (* better have an attribute *)
	case a of NONE => raise (SyntaxError "internal error?")
                | (SOME a) => a

      val d = case fetch dset of (DSET d) => d
                               | EMPTY => raise EmptyDataset

      val (v as {data, multichan, null, inherits, dielistener}) = 
	SyncVar.mTake d

      val ({name, modtime, state, 
	    attrs, subdataset, 
	    iattrs, isubdataset}, others) =
	case List.partition (fn (e : entry) => 
			     name e = ename) data of
	  (nil, others) => raise NotPermitted
	| ([e], others) => (e, others)
	| _ => raise DoubleEntry

    in
      raise NotImplemented
    end

end	
