functor EntryFun(structure Acl' : ACL
		 structure AcapTime' : ACAPTIME) :> ENTRY 
  where type Acl.acl = Acl'.acl 
    and type AcapTime.acaptime = AcapTime'.acaptime =
struct
  structure Acl = Acl'
  structure AcapTime = AcapTime'
  structure V = Value

  type metadata = {acl : Acl.acl option, attribute : string,
		   size : int list, value : V.value}

  type storedata = {acl : Acl.acl option, attribute : string, 
		    value : V.value option}

  type attribute = metadata

  type entry = {name : string, shortname : string,
		modtime : AcapTime.acaptime, 
		attrs : attribute list}

  val nullMeta = {acl=NONE, attribute="", size=nil, value=V.Nil}

  fun makeNull attr = {acl=NONE, attribute=attr, size=nil, value=V.Nil}

  (* NEED TO DO THIS!!! *)
  exception EnforcedValidation of string

  exception Modified of AcapTime.acaptime

  exception SyntaxError of string

  exception DoubleAttribute

  fun computeSize (V.Single s) = [String.size s]
    | computeSize (V.List l) = map String.size l
    | computeSize V.Nil = nil
    | computeSize V.Default = nil

  exception ChangeName of string
  exception DeleteEntry

  fun storeToMeta (({acl, attribute, value} : storedata),
                  ({acl=oldAcl, attribute=oldAttr, 
		    size=oldSize, value=oldValue} : metadata)) =
    let 
      val _ = if attribute = "modtime" then 
	         raise (EnforcedValidation "modtime")
	      else ()

      (* are we attempting to delete or rename this entry? *)
      val _ = if attribute = "entry" then
	      case value of
		NONE => ()
	      | (SOME (V.List _)) => raise 
		  (EnforcedValidation "entry")
	      | (SOME (V.Single n)) => raise (ChangeName n)
	      | (SOME (V.Default)) => raise (DeleteEntry)
	      | (SOME (V.Nil)) => raise (DeleteEntry)
	      else ()

      val acl = if isSome acl then acl else oldAcl
      val (value, size) = if isSome value then (valOf value, 
						computeSize (valOf value))
			  else (oldValue, oldSize)
    in
      {acl=acl, attribute=attribute, size=size, value=value}
    end

  (* given a full path, return the last component *)
  fun getshort name = List.last (String.fields (fn c => c = #"/") name)

  (* creates an entry, given it's name *)
  (* val create : string -> storedata list -> entry *)
  fun create name sd = {name=name, 
			shortname=getshort name,
			modtime=AcapTime.now (), 
			attrs=List.map 
			(fn x => storeToMeta (x, nullMeta)) sd}
    handle _ => raise (SyntaxError "this create made no sense")

  (* stores a list of attributes in this entry; this is where
     enforced validation takes place. returns NONE if this entry
     was deleted *)

  fun store' entry nil = (SOME entry)
    | store' {attrs,name,modtime,shortname} 
             ((sd as {attribute=tochange,...})::l) = 
    let
      val (a, others) = (List.partition 
			 (fn ({attribute=a,...}) => (a = tochange)) attrs)

    in
      (store' {attrs=(case a of nil => storeToMeta(sd, nullMeta)
                      | [a] => storeToMeta(sd, a)
		      | _ => raise DoubleAttribute)::others,
	       name=name, modtime=modtime, shortname=shortname} l) handle 
	ChangeName n => store' {attrs=attrs,name=n,modtime=modtime,
				shortname=getshort n} l
      | DeleteEntry => if List.null l then NONE
		       else raise (SyntaxError "trying to delete and store")
    end

  (* val store : entry -> storedata list -> time option -> entry option *)
  fun store (entry as {modtime,name,attrs,shortname}) sd (SOME time) =
    (case (AcapTime.compare (modtime, time)) of
       GREATER => raise (Modified modtime)
     | _ => store' entry sd)
    | store {shortname,name,modtime,attrs} sd NONE = 
       store' {name=name,modtime=AcapTime.now (),
	       attrs=attrs,shortname=shortname} sd

  (* get the entry's name *)
  (* val name : entry -> name *)
  fun name ({shortname,...}:entry) = shortname

  (* get the entry's path *)
  (* val path : entry -> name *)
  fun path ({name,...}:entry) = name

  (* fetch and search have to return "entry" and "modtime" attributes! *)
  fun entry' ({shortname,...}:entry) =
    {acl=NONE, attribute="entry", size=[(String.size shortname)],
     value = V.Single shortname}

  fun modtime' ({modtime,...}:entry) = 
    let
      val v = AcapTime.toValue modtime
    in
      {acl=NONE, attribute="modtime", size=Value.size v, value=v}
    end

  (* gets the attributes named *)
  (* val fetch : entry -> string list -> attribute list *)
  fun fetch (e as {attrs, ...}) l = 
    let
      fun fetch' nil out = List.rev out
	| fetch' (an::l) out =
	let
	  val attr = if an = "entry" then SOME (entry' e)
		     else if an = "modtime" then SOME (modtime' e)
			  else List.find (fn ({attribute,...}:attribute) =>
					  an = attribute) attrs

	  val attr = getOpt (attr, makeNull an)
	in
	  fetch' l (attr::out)
	end
    in
      fetch' l nil
    end

  (* gets the matching attributes *)
  (* val search : entry -> (string -> bool) -> attribute list *)
  fun search (e as {attrs,...}) f =
    let
      val attrs = List.filter (fn ({attribute,...}) => f attribute) attrs

      (* now we figure out if "entry" or "modtime" match *)
      val attrs = if f "entry" then (entry' e)::attrs else attrs

      val attrs = if f "modtime" then (modtime' e)::attrs
		  else attrs
    in
      attrs
    end

  (* return the time of last modification *)
  (* val gettime : entry -> acaptime *)
  fun gettime ({modtime,...} : entry) = modtime      

end