(* sasl.sml
 *
 * functor producing structures with SML/NJ SASL interface
 *)

functor C_SASL (structure C : C_CALLS) : C_SASL =
struct
  open C

  exception SaslError of string

  type connection = caddr

  type security_flags = Word32.word

  datatype response = 
    OK of string
  | CONTINUE of string

  fun ptr c = Cptr (Caddr c)

  val _ = print "building the SASL C interface..."

  val sasl_done' = registerAutoFreeCFn("sasl_done", [], CvoidT)

  fun sasl_done () = ignore (sasl_done' [Cvoid])

  val sasl_dispose' = registerAutoFreeCFn("sasl_dispose", 
					  [CptrT CaddrT],
					  CvoidT)

  fun sasl_dispose conn = ignore (sasl_dispose' [ptr conn])

  val sasl_usererr' = registerAutoFreeCFn("sasl_usererr", [CintT], CintT)

  fun sasl_usererr i = let val Cint i = sasl_usererr' [Cint i] in i end

  val sasl_errstring' = registerAutoFreeCFn("sasl_errstring", 
					    [CintT, 
					     CaddrT, 
					     CptrT CaddrT],
					    CstringT)
							       
  fun sasl_errstring i = 
    let val (Cstring s) = sasl_errstring' [Cint i, Caddr NULL, 
					   Cptr (Caddr NULL)] 
    in s end

  (* given a SASL return value, raise exception or not *)
  fun retExn (0w0 | 0w1) = ()
    | retExn (x : Word32.word) = raise (SaslError (sasl_errstring x))

  val sasl_getprop' = registerCFn("sasl_getprop",
				  [CaddrT, CintT, CptrT CaddrT],
				  CintT)

  fun sasl_get_username conn = 
    let
      val (Cint ret, l as [caddr]) = 
	sasl_getprop' [Caddr conn, Cint (Word32.fromInt 0), Cptr (Caddr NULL)]
	
      val (Cptr (Cstring s)) = datumCtoML (CptrT CstringT) caddr
      val (Cptr (Caddr caddr)) = datumCtoML (CptrT CaddrT) caddr
      val l = caddr::l
    in
      app free l; (* free all memory *)
      retExn ret;
      s
    end

  (* need to create structures to pass to this *)
  val sasl_setprop' = registerCFn("sasl_setprop",
				  [CaddrT, CintT, CaddrT],
				  CintT)
    
    
  fun sasl_build_secflags () = (0w0 : Word32.word)

  val SASL_SEC_PROPS = (Word32.fromInt 101)

  fun sasl_set_security_properties conn (min, max, maxbuf, secflags) =
    let
      val (caddr, l) = datumMLtoC (CptrT (CstructT [CintT, CintT, 
						    CintT, CintT,
						    CptrT CaddrT, 
						    CptrT CaddrT]))
	                          (Cptr (Cstruct [Cint (Word32.fromInt min),
						  Cint (Word32.fromInt max),
						  Cint (Word32.fromInt maxbuf),
						  Cint (secflags), 
						  Cptr (Caddr NULL),
						  Cptr (Caddr NULL)]))

      val (Cint ret, l2) = sasl_setprop' [Caddr conn, Cint SASL_SEC_PROPS, 
					  Caddr caddr]
    in
      app free l;
      app free l2;
      retExn ret
    end

  val sasl_idle' = registerAutoFreeCFn("sasl_idle", [CaddrT], CintT)

  fun sasl_idle conn =
    let val (Cint ret) = (sasl_idle' [Caddr conn]) in
      retExn ret
    end
        
  (* client API *)
  val sasl_client_init' = registerCFn("sasl_client_init",
				      [CptrT CaddrT],
				      CintT)

  val sasl_client_new' = registerCFn("sasl_client_new",
				     [CptrT CstringT,
				      CptrT CstringT,
   				      CaddrT, (* should be a callback *)
				      CintT,
				      CptrT CaddrT],
				     CintT)
    
  val sasl_client_start' = registerCFn("sasl_client_start",
				       [CaddrT, (* connection *)
					CstringT, (* mech list *)
					CaddrT, (* secret *)
					CptrT CaddrT, (* prompt_need *)
					CptrT CaddrT, (* clientout *)
					CptrT CintT,
					CptrT CstringT], (* mech *)
				       CintT)

  val sasl_client_step' = registerCFn("sasl_client_step",
				      [CaddrT,
				       CaddrT, (* server in *)
				       CintT,
				       CptrT CaddrT, (* prompt_need *)
				       CptrT CaddrT, (* clientout *)
				       CptrT CintT],
				      CintT)

  val sasl_client_auth' = registerCFn("sasl_client_auth",
				      [CaddrT,
				       CstringT,
				       CstringT,
				       CintT, (* passlen *)
				       CaddrT, (* prompts *)
				       CptrT CaddrT], (* keepcopy *)
				      CintT)

  val sasl_free_secret' = registerCFn("sasl_free_secret",
				      [CptrT CaddrT],
				      CvoidT)

  (* server API *)
  val sasl_server_init' = registerAutoFreeCFn("sasl_server_init",
				      [CptrT CaddrT,
				       CstringT], (* appname *)
				      CintT)

  fun sasl_server_init appname = 
    let val (Cint ret) = sasl_server_init' [Cptr (Caddr NULL), Cstring appname]
    in retExn ret end
    
  (* all optional strings must be CaddrT's *)
  val sasl_server_new' = registerCFn("sasl_server_new",
				     [CstringT, (* service *)
				      CaddrT, (* serverFQDN *)
				      CaddrT, (* user_realm *)
				      CaddrT, (* callbacks *)
				      CintT, (* secflags *)
				      CptrT CaddrT], (* conn *)
				     CintT)


  (* sasl_server_new : string -> security_flags -> connection *)
  fun sasl_server_new service secflags =
    let
      val (Cint ret, l as [_, caddr]) =
	sasl_server_new' [Cstring service, Caddr NULL, Caddr NULL, Caddr NULL,
			  Cint secflags, Cptr (Caddr NULL)]

      val (Cptr (Caddr caddr)) = datumCtoML (CptrT CaddrT) caddr
    in
      app free l;
      retExn ret;
      caddr
    end

  val sasl_listmech' = registerCFn("sasl_listmech",
				   [CaddrT, (* conn *)
				    CaddrT, (* user *)
				    CstringT, (* prefix *)
				    CstringT, (* sep *)
				    CstringT, (* suffix *)
				    CptrT CaddrT, (* result *)
				    CaddrT, (* len *)
				    CaddrT], (* count *)
				   CintT)

  (* sasl_listmech : conn -> string *)
  fun sasl_listmech (prefix, sep, suffix) conn =
    let
      val (Cint ret, l as [_, _, _, res]) =
	sasl_listmech' [Caddr conn, Caddr NULL, Cstring prefix, Cstring sep,
			Cstring suffix, Cptr (Caddr NULL), 
			Caddr NULL, Caddr NULL]

      val (Cptr (Cstring s)) = datumCtoML (CptrT CstringT) res

      val (Cptr (Caddr caddr)) = datumCtoML (CptrT CaddrT) res

      val l = caddr::l
    in
      app free l;
      retExn ret;
      s
    end

  fun makevec s = datumMLtoC (CptrT (CvectorT (size s, CcharT)))
    (Cptr (Cvector (Vector.fromList (List.map Cchar (String.explode s)))))

  fun makestring' (0, ptr, acc) = CharVector.fromList (List.rev acc)
    | makestring' (n, ptr, acc) = 
    let
      val (Cptr (Cchar c)) = datumCtoML (CptrT CcharT) ptr
    in
      makestring' (n-1, index(ptr, 1), c::acc)
    end

  fun makestring len ptr = 
    let
      val (Cptr (Caddr deref)) = datumCtoML (CptrT CaddrT) ptr
    in 
      makestring' (len, deref, nil)
    end

  val sasl_server_start' = registerCFn("sasl_server_start",
				       [CaddrT, (* conn *)
					CstringT, (* mech *)
					CaddrT, (* client in *)
					CintT, (* inlen *)
					CptrT CaddrT, (* out *)
					CptrT CintT, (* outlen *)
					CaddrT], (* errstr *)
				       CintT)

  fun sasl_server_start conn mech clientin =
    let
      val (cin, l) = 
	case clientin of
	  (SOME clientin) => makevec clientin
	| NONE => (NULL, nil)
			   
      val cinlen = 
	case clientin of
	  (SOME clientin) => Cint (Word32.fromInt (size clientin))
	| NONE => Cint (0w0)

      val (Cint ret, l2 as [_, out, outlen]) =
	sasl_server_start' [Caddr conn, Cstring mech, Caddr cin, cinlen,
			    Cptr (Caddr NULL), Cptr (Cint (0w0 : Word32.word)),
			    (Caddr NULL)]

      val (Cptr (Cint outlen)) = datumCtoML (CptrT CintT) outlen
	handle NotAPtr _ => (Cptr (Cint 0w0))
	  
      val outlen = Word32.toInt outlen

      val _ = print ("want to send " ^ (Int.toString outlen) ^ " bytes\n")

      val s = makestring outlen out
	handle NotAPtr _ => ""

      val (Cptr (Caddr caddr)) = datumCtoML (CptrT (CaddrT)) out
	handle NotAPtr _ => (Cptr (Caddr NULL))
    in
      app free l;
      app free l2;
      if (caddr <> NULL) then free caddr else ();
      retExn ret;
      case ret of
	0w0 => OK s
      | _ => CONTINUE s
    end


  val sasl_server_step' = registerCFn("sasl_server_step",
				      [CaddrT, (* conn *)
				       CaddrT, (* in *)
				       CintT, (* inlen *)
				       CptrT CaddrT, (* out *)
				       CptrT CintT, (* outlen *)
				       CaddrT], (* errstr *)
				      CintT)

  fun sasl_server_step conn clientin =
    let
      val (cin, l) = makevec clientin
			   
      val cinlen = Cint (Word32.fromInt (size clientin))

      val (Cint ret, l2 as [out, outlen]) =
	sasl_server_step' [Caddr conn, Caddr cin, cinlen,
			   Cptr (Caddr NULL), Cptr (Cint 0w0),
			   (Caddr NULL)]

      val (Cptr (Cint outlen)) = datumCtoML (CptrT CintT) outlen
	handle NotAPtr _ => (Cptr (Cint 0w0))

      val s = makestring (Word32.toInt outlen) out
	handle NotAPtr _ => ""

      val (Cptr (Caddr caddr)) = datumCtoML (CptrT (CaddrT)) out
	handle NotAPtr _ => (Cptr (Caddr NULL))
    in
      app free l;
      app free l2;
      if (caddr <> NULL) then free caddr else ();
      retExn ret;
      case ret of
	0w0 => OK s
      | _ => CONTINUE s
    end

  val sasl_checkpass' = registerCFn("sasl_checkpass",
				    [CstringT, (* user *)
				     CintT, (* userlen *)
				     CstringT, (* pass *)
				     CintT, (* passlen *)
				     CaddrT], (* errstr *)
				    CintT)

  fun sasl_checkpass (user, pass) =
    let
      val (ulen) = Word32.fromInt (size user)
      val (user, l) = makevec user
				 
      val (plen) = Word32.fromInt (size pass)
      val (pass, l2) = makevec pass

      val (Cint ret, l3 as []) =
	sasl_checkpass' [Caddr user, (Cint ulen), Caddr pass, (Cint plen), 
			 Caddr NULL]
    in
      app free l;
      app free l2;
      app free l3;
      if ret = 0w0 then true
      else false
    end

(* not implemented!
  val sasl_user_exists' = registerCFn("sasl_user_exists",
				      [CstringT, (* service *)
				       CaddrT, (* realm *)
				       CstringT], (* user *)
				      CintT)
*)
  val sasl_setpass' = registerCFn("sasl_setpass",
				  [CaddrT, (* conn *)
				   CstringT, (* user *)
				   CstringT, (* pass *)
				   CintT, (* passlen *)
				   CintT, (* flags *)
				   CptrT CstringT], (* errstr *)
				  CintT)

  (* security layer *)
  val sasl_encode' = registerCFn("sasl_encode", 
				 [CaddrT, (* conn *)
				  CaddrT, (* input *)
				  CintT, (* inlen *)
				  CptrT CaddrT, (* out *)
				  CptrT CintT], (* outlen *)
				 CintT)

  val sasl_decode' = registerCFn("sasl_decode",
				 [CaddrT,
				  CaddrT,
				  CintT,
				  CptrT CaddrT,
				  CptrT CintT],
				 CintT)
    
  val _ = print "done.\n"

end
